#include <SdFat.h>
#include <EEPROM.h>

#include <SPI.h>
#include <SSD1306_text.h>
//#include "SSD1306Ascii.h"
//#include "SSD1306AsciiSpi.h"


#define DEBUG
#ifdef DEBUG
   #define DEBUGLN(X) Serial.println(F(X));
   #define DEBUGP(X) Serial.print(X);
#else
   #define DEBUGLN(X) 
   #define DEBUGP(X) 
#endif

//the directional keys on the arduboy
// i'm using inside/outside buttons because they've named a/b opposite to the NES convention
// same pins as arduboy for sanity
const byte BTN_UP    = A0;
const byte BTN_DOWN  = A3;
const byte BTN_LEFT  = A2;
const byte BTN_RIGHT = A1;
const byte BTN_IN    = 7;
const byte BTN_OUT   = 8;
const byte BTN_MENU  = A7; //6; //dedicated "menu" button, wakes up ICSP core and shuts down arduboy core

//SD card detect pin
const byte PIN_SD_CD = A6; //9;

//bit banged SPI pins (Atmega328p)
const byte MSPIM_SCK = 2;  // port D bit 2  =>arduboy-sck (15)
const byte MSPIM_SS  = 3;  // port D bit 3  =>arduboy-rst (rst)
const byte BB_MISO   = 4;  // port D bit 4  =>arduboy miso(14)
const byte BB_MOSI   = 5;  // port D bit 5  =>arduboy mosi(16)

#define BB_MISO_PORT PIND
#define BB_MOSI_PORT PORTD
#define BB_SCK_PORT PORTD
const byte BB_SCK_BIT = 2;
const byte BB_MISO_BIT = 4; //
const byte BB_MOSI_BIT = 5;

//Display
const byte PIN_OLED_RES = 9;
const byte PIN_OLED_CS  = A4; //A4, 18, PC4
const byte PIN_OLED_DC  = A5; //A5, 19, PC5
SSD1306_text oled(PIN_OLED_DC, PIN_OLED_RES, PIN_OLED_CS);
//SSD1306AsciiSpi oled;
const byte PIN_OLED_PLEX = 6;

//oled SCL = SCK => pro mini 13
//oled SDA = MOSI => pro mini 11

const byte RESET = MSPIM_SS; // target board reset pin
const uint8_t chipSelect = SS; // SD chip select pin

#define ANALOG_THRESHOLD 250 //when reading an analog pin as digital, what's the tipping point?


long currButtonState = 0;
long prevButtonState = 0;


// status "messages"
typedef enum 
{ 
   MSG_NO_SD_CARD,          // cannot open SD card
   MSG_SD_CARD_INIT_FAILED,  // egh
   MSG_CANNOT_OPEN_FILE,    // canoot open file 'wantedFile' (above)
   MSG_LINE_TOO_LONG,       // line on disk too long to read
   MSG_LINE_TOO_SHORT,      // line too short to be valid
   MSG_LINE_DOES_NOT_START_WITH_COLON,  // line does not start with a colon
   MSG_INVALID_HEX_DIGITS,  // invalid hex where there should be hex
   MSG_BAD_SUMCHECK,        // line fails sumcheck
   MSG_LINE_NOT_EXPECTED_LENGTH,  // record not length expected
   MSG_UNKNOWN_RECORD_TYPE,  // record type not known
   MSG_NO_END_OF_FILE_RECORD,  // no 'end of file' at end of file
   MSG_FILE_TOO_LARGE_FOR_FLASH,  // file will not fit into flash
   
   MSG_CANNOT_ENTER_PROGRAMMING_MODE,  // cannot program target chip
   MSG_NO_BOOTLOADER_FUSE,             // chip does not have bootloader
   MSG_CANNOT_FIND_SIGNATURE,      // cannot find chip signature
   MSG_UNRECOGNIZED_SIGNATURE,     // signature not known
   MSG_BAD_START_ADDRESS,          // file start address invalid
   MSG_VERIFICATION_ERROR,         // verification error after programming
   MSG_FLASHED_OK,                 // flashed OK
 } msgType;

const unsigned int ENTER_PROGRAMMING_ATTEMPTS = 10;





const unsigned long NO_PAGE = 0xFFFFFFFF; //?


SdFat sd;        // file system object
char filename[13]; // buffer for storing filename to load

ifstream listFileStream;
int listLineIdx = 0;          // what line is currently selected?
char listLineBuff[45];        // the currently selected line from list.txt, copied to memory
const byte listLineSize = 43; // how long is each line in list.txt
int listLineCount = 0;        // how many lines are in list.txt?




// copy of fuses/lock bytes found for this processor
byte fuses[5];  

// meaning of bytes in above array
enum
{
   lowFuse,
   highFuse,
   extFuse,
   lockByte,
   calibrationByte
};


// actions to take
enum 
{
   checkFile,
   verifyFlash,
   writeToFlash,
};


// structure to hold signature and other relevant data about each chip
typedef struct {
   byte sig [3];
   const char * desc;
   unsigned long flashSize;
   unsigned int baseBootSize;
   unsigned long pageSize;     // bytes
   byte fuseWithBootloaderSize;  // ie. one of: lowFuse, highFuse, extFuse
   byte timedWrites;    // if pollUntilReady won't work by polling the chip
} signatureType;

const unsigned long kb = 1024;
const byte NO_FUSE = 0xFF;

// see Atmega datasheets
const signatureType signatures [] PROGMEM = 
{
   //     signature        description   flash size   bootloader  flash  fuse
   //                                                     size    page    to
   //                                                             size   change
   
   // Attiny84 family
   { { 0x1E, 0x91, 0x0B }, "ATtiny24",   2 * kb,           0,   32,   NO_FUSE },
   { { 0x1E, 0x92, 0x07 }, "ATtiny44",   4 * kb,           0,   64,   NO_FUSE },
   { { 0x1E, 0x93, 0x0C }, "ATtiny84",   8 * kb,           0,   64,   NO_FUSE },
   
   // Attiny85 family
   { { 0x1E, 0x91, 0x08 }, "ATtiny25",   2 * kb,           0,   32,   NO_FUSE },
   { { 0x1E, 0x92, 0x06 }, "ATtiny45",   4 * kb,           0,   64,   NO_FUSE },
   { { 0x1E, 0x93, 0x0B }, "ATtiny85",   8 * kb,           0,   64,   NO_FUSE },
   
   // Atmega328 family
   { { 0x1E, 0x92, 0x0A }, "ATmega48PA",   4 * kb,         0,    64,  NO_FUSE },
   { { 0x1E, 0x93, 0x0F }, "ATmega88PA",   8 * kb,       256,   128,  extFuse },
   { { 0x1E, 0x94, 0x0B }, "ATmega168PA", 16 * kb,       256,   128,  extFuse },
   { { 0x1E, 0x95, 0x0F }, "ATmega328P",  32 * kb,       512,   128,  highFuse },
   
   // Atmega644 family
   { { 0x1E, 0x94, 0x0A }, "ATmega164P",   16 * kb,      256,   128,  highFuse },
   { { 0x1E, 0x95, 0x08 }, "ATmega324P",   32 * kb,      512,   128,  highFuse },
   { { 0x1E, 0x96, 0x0A }, "ATmega644P",   64 * kb,   1 * kb,   256,  highFuse },
   
   // Atmega2560 family
   { { 0x1E, 0x96, 0x08 }, "ATmega640",    64 * kb,   1 * kb,   256,  highFuse },
   { { 0x1E, 0x97, 0x03 }, "ATmega1280",  128 * kb,   1 * kb,   256,  highFuse },
   { { 0x1E, 0x97, 0x04 }, "ATmega1281",  128 * kb,   1 * kb,   256,  highFuse },
   { { 0x1E, 0x98, 0x01 }, "ATmega2560",  256 * kb,   1 * kb,   256,  highFuse },
      
   { { 0x1E, 0x98, 0x02 }, "ATmega2561",  256 * kb,   1 * kb,   256,  highFuse },
   
   // AT90USB family
   { { 0x1E, 0x93, 0x82 }, "At90USB82",    8 * kb,       512,   128,  highFuse },
   { { 0x1E, 0x94, 0x82 }, "At90USB162",  16 * kb,       512,   128,  highFuse },
   
   // Atmega32U2 family
   { { 0x1E, 0x93, 0x89 }, "ATmega8U2",    8 * kb,       512,   128,  highFuse  },
   { { 0x1E, 0x94, 0x89 }, "ATmega16U2",  16 * kb,       512,   128,  highFuse  },
   { { 0x1E, 0x95, 0x8A }, "ATmega32U2",  32 * kb,       512,   128,  highFuse  },
   
   // Atmega32U4 family -  (datasheet is wrong about flash page size being 128 words)
   { { 0x1E, 0x94, 0x88 }, "ATmega16U4",  16 * kb,       512,   128,  highFuse },
   { { 0x1E, 0x95, 0x87 }, "ATmega32U4",  32 * kb,       512,   128,  highFuse },
   
   // ATmega1284P family
   { { 0x1E, 0x97, 0x05 }, "ATmega1284P", 128 * kb,   1 * kb,   256,  highFuse  },
   
   // ATtiny4313 family
   { { 0x1E, 0x91, 0x0A }, "ATtiny2313A",   2 * kb,        0,    32,  NO_FUSE  },
   { { 0x1E, 0x92, 0x0D }, "ATtiny4313",    4 * kb,        0,    64,  NO_FUSE  },
   
   // ATtiny13 family
   { { 0x1E, 0x90, 0x07 }, "ATtiny13A",     1 * kb,        0,    32,  NO_FUSE },
   
   // Atmega8A family
   { { 0x1E, 0x93, 0x07 }, "ATmega8A",      8 * kb,      256,    64,  highFuse, true },
   
   // ATmega64rfr2 family
   { { 0x1E, 0xA6, 0x02 }, "ATmega64rfr2",  256 * kb, 1 * kb,   256, highFuse },
   { { 0x1E, 0xA7, 0x02 }, "ATmega128rfr2", 256 * kb, 1 * kb,   256, highFuse },
   { { 0x1E, 0xA8, 0x02 }, "ATmega256rfr2", 256 * kb, 1 * kb,   256, highFuse },

};  // end of signatures

// if signature found in above table, this is its index
int foundSig = -1;
byte lastAddressMSB = 0;
// copy of current signature entry for matching processor
signatureType currentSignature;

       
// number of items in an array
#define NUMITEMS(arg) ((unsigned int) (sizeof (arg) / sizeof (arg [0])))


unsigned long pagesize;
unsigned long pagemask;
unsigned long oldPage;
unsigned int progressBarCount;


bool gotEndOfFile;
unsigned long extendedAddress;

unsigned long lowestAddress;
unsigned long highestAddress;
unsigned long bytesWritten;
unsigned int lineCount;


// count errors
unsigned int errors;


// programming commands to send via SPI to the chip
enum 
{
   progamEnable = 0xAC,
   
   // writes are preceded by progamEnable
   chipErase             = 0x80,
   writeLockByte         = 0xE0,
   writeLowFuseByte      = 0xA0,
   writeHighFuseByte     = 0xA8,
   writeExtendedFuseByte = 0xA4,
   
   pollReady = 0xF0,
   
   programAcknowledge = 0x53,

   
   readSignatureByte   = 0x30,
   readCalibrationByte = 0x38,
   
   readLowFuseByte     = 0x50,       
   readLowFuseByteArg2 = 0x00,
   
   readExtendedFuseByte     = 0x50,  
   readExtendedFuseByteArg2 = 0x08,
   
   readHighFuseByte     = 0x58,      
   readHighFuseByteArg2 = 0x08,  
   
   
   readLockByte    = 0x58,         
   readLockByteArg2 = 0x00,  
   
   readProgramMemory       = 0x20,  
   writeProgramMemory      = 0x4C,
   loadExtendedAddressByte = 0x4D,
   loadProgramMemory       = 0x40,

   //egh
   readEEPROMMemory = 0xA0,
   writeEEPROMMemory = 0xC0,
   
};  


// which program instruction writes which fuse
const byte fuseCommands [4] = { writeLowFuseByte, writeHighFuseByte, writeExtendedFuseByte, writeLockByte };

// types of record in .hex file
enum 
{
    hexDataRecord,                   // 00
    hexEndOfFile,                    // 01
    hexExtendedSegmentAddressRecord, // 02
    hexStartSegmentAddressRecord,    // 03
    hexExtendedLinearAddressRecord,  // 04
    hexStartLinearAddressRecord      // 05
};


void ShowMessage(const byte which)
{
   //TODO: make message display on the screen
   #ifdef DEBUG
   switch (which)
   {
      // problems with SD card or finding the file
      case MSG_NO_SD_CARD:                      Serial.println(F("MSG_NO_SD_CARD")); break;
      case MSG_SD_CARD_INIT_FAILED:             Serial.println(F("MSG_SD_CARD_INIT_FAILED")); break;
      case MSG_CANNOT_OPEN_FILE:                Serial.println(F("MSG_CANNOT_OPEN_FILE")); break;
      
      // problems reading the .hex file
      case MSG_LINE_TOO_LONG:                   Serial.println(F("MSG_LINE_TOO_LONG")); break;
      case MSG_LINE_TOO_SHORT:                  Serial.println(F("MSG_LINE_TOO_SHORT")); break;
      case MSG_LINE_DOES_NOT_START_WITH_COLON:  Serial.println(F("MSG_LINE_DOES_NOT_START_WITH_COLON")); break;
      case MSG_INVALID_HEX_DIGITS:              Serial.println(F("MSG_INVALID_HE")); break;
      case MSG_BAD_SUMCHECK:                    Serial.println(F("MSG_BAD_SUMCHECK")); break;
      case MSG_LINE_NOT_EXPECTED_LENGTH:        Serial.println(F("MSG_LINE_NOT_EXPECTED_LENGTH")); break;
      case MSG_UNKNOWN_RECORD_TYPE:             Serial.println(F("MSG_UNKNOWN_RECORD_TYPE")); break;
      case MSG_NO_END_OF_FILE_RECORD:           Serial.println(F("MSG_NO_END_OF_FILE_RECORD")); break;
      
      // problems with the file contents
      case MSG_FILE_TOO_LARGE_FOR_FLASH:        Serial.println(F("MSG_FILE_TOO_LARGE_FOR_FLASH")); break;
      
      // problems programming the chip
      case MSG_CANNOT_ENTER_PROGRAMMING_MODE:  Serial.println(F("MSG_CANNOT_ENTER_PROGRAMMING_MODE")); break;
      case MSG_NO_BOOTLOADER_FUSE:             Serial.println(F("MSG_NO_BOOTLOADER_FUSE")); break;
      case MSG_CANNOT_FIND_SIGNATURE:          Serial.println(F("MSG_CANNOT_FIND_SIGNATURE")); break;
      case MSG_UNRECOGNIZED_SIGNATURE:         Serial.println(F("MSG_UNRECOGNIZED_SIGNATURE")); break;
      case MSG_BAD_START_ADDRESS:              Serial.println(F("MSG_BAD_START_ADDRESS")); break;
      case MSG_VERIFICATION_ERROR:             Serial.println(F("MSG_VERIFICATION_ERROR")); break;
      case MSG_FLASHED_OK:                     Serial.println(F("MSG_FLASHED_OK")); break;

      // unknown error
      default:                                 Serial.println(F("Default message")); break;   
   }
   #endif
   
   switch (which)
   {
      // problems with SD card or finding the file
      case MSG_NO_SD_CARD:                      break;
      case MSG_SD_CARD_INIT_FAILED:             break;
      case MSG_CANNOT_OPEN_FILE:                break;
      
      // problems reading the .hex file
      case MSG_LINE_TOO_LONG:                   break;
      case MSG_LINE_TOO_SHORT:                  break;
      case MSG_LINE_DOES_NOT_START_WITH_COLON:  break;
      case MSG_INVALID_HEX_DIGITS:              break;
      case MSG_BAD_SUMCHECK:                    break;
      case MSG_LINE_NOT_EXPECTED_LENGTH:        break;
      case MSG_UNKNOWN_RECORD_TYPE:             break;
      case MSG_NO_END_OF_FILE_RECORD:           break;
      
      // problems with the file contents
      case MSG_FILE_TOO_LARGE_FOR_FLASH:        break;
      
      // problems programming the chip
      case MSG_CANNOT_ENTER_PROGRAMMING_MODE:  break;
      case MSG_NO_BOOTLOADER_FUSE:             break;
      case MSG_CANNOT_FIND_SIGNATURE:          break;
      case MSG_UNRECOGNIZED_SIGNATURE:         break;
      case MSG_BAD_START_ADDRESS:              break;
      case MSG_VERIFICATION_ERROR:             break;
      case MSG_FLASHED_OK:                     break;

      // unknown error
      default:                                 break;   
   }
}


// Bit Banged SPI transfer
inline byte BB_SPITransfer(byte c)
{       
   byte bit;

   //EGH: hardware spi
   // the issue is that even connecting the two SCK pins on the arduinos causes the sd card to have read errors
   // and other problems
   //take the SS pin low to select the chip:
   //digitalWrite(RESET, LOW);
   //c = SPI.transfer(c);
   //digitalWrite(RESET, HIGH);
   
   for (bit = 0; bit < 8; bit++) 
   {
      // write MOSI on falling edge of previous clock
      if (c & 0x80)
         BB_MOSI_PORT |= bit(BB_MOSI_BIT);
      else
         BB_MOSI_PORT &= ~bit(BB_MOSI_BIT);
         
      c <<= 1;
      
      // read MISO
      c |= (BB_MISO_PORT & bit(BB_MISO_BIT)) != 0;
      
      // clock high
      BB_SCK_PORT |= bit(BB_SCK_BIT);
      
      // delay between rise and fall of clock
      __asm__("nop\n\t");
      
      // clock low
      BB_SCK_PORT &= ~bit(BB_SCK_BIT);
      
      // delay between rise and fall of clock
      __asm__("nop\n\t");
   }
   
   return c;
}  


// execute one programming instruction ... b1 is command, b2, b3, b4 are arguments
//  processor may return a result on the 4th transfer, this is returned.
byte program(const byte b1, const byte b2 = 0, const byte b3 = 0, const byte b4 = 0)
{
   noInterrupts();
   BB_SPITransfer(b1);  
   BB_SPITransfer(b2);  
   BB_SPITransfer(b3);  
   byte b = BB_SPITransfer(b4);  
   interrupts();
   return b;
} 


// read a byte from flash memory
byte readFlash(unsigned long addr)
{
   // set if high byte wanted
   byte high = (addr & 1) ? 0x08 : 0;  

   // turn into word address
   addr >>= 1;  
   
   // set the extended (most significant) address byte if necessary
   byte MSB = (addr >> 16) & 0xFF;
   if (MSB != lastAddressMSB)
   {
      program(loadExtendedAddressByte, 0, MSB); 
      lastAddressMSB = MSB;
   }  
   
   return program(readProgramMemory | high, highByte(addr), lowByte(addr));
} 


// write a byte to the flash memory buffer (ready for committing)
void writeFlash(unsigned long addr, const byte data)
{
   // set if high byte wanted
   byte high = (addr & 1) ? 0x08 : 0;  

   // turn into word address
   addr >>= 1;  
   
   program(loadProgramMemory | high, 0, lowByte(addr), data);
}  


// convert two hex characters into a byte
//    returns true if error, false if OK
bool hexConv(const char * (& pStr), byte & b)
{

   if (!isxdigit(pStr [0]) || !isxdigit(pStr [1]))
   {
      ShowMessage(MSG_INVALID_HEX_DIGITS);
      return true;
   }
   
   b = *pStr++ - '0';
   if(b > 9)
      b -= 7;
   
   // high-order nybble
   b <<= 4;
   
   byte b1 = *pStr++ - '0';
   if(b1 > 9)
      b1 -= 7;
   
   b |= b1;
   
   return false;  // OK
}

// poll the target device until it is ready to be programmed
void pollUntilReady()
{
   if(currentSignature.timedWrites)
   {
      // at least 2 x WD_FLASH which is 4.5 mS
      delay(10);  
   }
   else
   {  
      while((program(pollReady) & 1) == 1)
      {  
         // wait till ready  
      }
   }
} 


void showProgress()
{
   //TODO: spin a thing on the screen

   DEBUGP("#");
   oled.writeWidth(".");
}  
  

// commit page to flash memory
void commitPage(unsigned long addr)
{
   addr >>= 1;  // turn into word address
   
   // set the extended (most significant) address byte if necessary
   byte MSB = (addr >> 16) & 0xFF;
   if (MSB != lastAddressMSB)
   {
      program(loadExtendedAddressByte, 0, MSB); 
      lastAddressMSB = MSB;
   }
   
   showProgress();
   
   program(writeProgramMemory, highByte(addr), lowByte(addr));
   pollUntilReady();    
}  


// write data to temporary buffer, ready for committing  
void writeData(const unsigned long addr, const byte * pData, const int length)
{
   // write each byte
   for (int i = 0; i < length; i++)
   {
      unsigned long thisPage = (addr + i) & pagemask;
      
      // page changed? commit old one
      if (thisPage != oldPage && oldPage != NO_PAGE)
         commitPage(oldPage);
         
      // now this is the current page
      oldPage = thisPage;
      
      // put byte into work buffer
      writeFlash(addr + i, pData [i]);
   }  
}

void verifyData(const unsigned long addr, const byte * pData, const int length)
{
   // check each byte
   for (int i = 0; i < length; i++)
   {
      unsigned long thisPage = (addr + i) & pagemask;
      
      // page changed? show progress
      if (thisPage != oldPage && oldPage != NO_PAGE)
         showProgress();
      
      // now this is the current page
      oldPage = thisPage;
      
      byte found = readFlash(addr + i);
      byte expected = pData[i];
      if (found != expected)
         errors++;
   }  
} 


/*
Line format:

  :nnaaaatt(data)ss
  
  Where:
  :      = a colon
  
  (All of below in hex format)
  
  nn     = length of data part
  aaaa   = address (eg. where to write data)
  tt     = transaction type
           00 = data
           01 = end of file
           02 = extended segment address (changes high-order byte of the address)
           03 = start segment address *
           04 = linear address *
           05 = start linear address *
  (data) = variable length data
  ss     = sumcheck

            * We don't use these
   
*/
// returns true if error, false if OK
bool processLine(const char * pLine, const byte action)
{
   if(*pLine++ != ':')
   {
      ShowMessage(MSG_LINE_DOES_NOT_START_WITH_COLON);
      return true;  // error
   } 
   
   const int maxHexData = 40;
   byte hexBuffer[maxHexData];
   int bytesInLine = 0;
   
   if(action == checkFile)
      if(lineCount++ % 40 == 0)
         showProgress();
   
   // convert entire line from ASCII into binary
   while(isxdigit(*pLine))
   {
      // too long?
      if(bytesInLine >= maxHexData)
      {
         ShowMessage(MSG_LINE_TOO_LONG);
         return true;
      }
      
      if(hexConv(pLine, hexBuffer[bytesInLine++]))
         return true;
   }
   
   if(bytesInLine < 5)
   {
      ShowMessage(MSG_LINE_TOO_SHORT);
      return true;  //error
   } 
   
   // sumcheck it
   /*
   byte sumCheck = 0;
   for (int i = 0; i < (bytesInLine - 1); i++)
   sumCheck += hexBuffer [i];
   
   // 2's complement
   sumCheck = ~sumCheck + 1;
   
   // check sumcheck
   if (sumCheck != hexBuffer [bytesInLine - 1])
   {
      ShowMessage (MSG_BAD_SUMCHECK);
      return true;
   }
   */
   
   // length of data (eg. how much to write to memory)
   byte len = hexBuffer[0];
   
   // the data length should be the number of bytes, less
   //   length / address (2) / transaction type / sumcheck
   if (len != (bytesInLine - 5))
   {
      ShowMessage(MSG_LINE_NOT_EXPECTED_LENGTH);
      return true; //error
   }
   
   // two bytes of address
   unsigned long addrH = hexBuffer[1];
   unsigned long addrL = hexBuffer[2];
   
   unsigned long addr = addrL | (addrH << 8);
   
   byte recType = hexBuffer[3];
   
   switch(recType)
   {
      // stuff to be written to memory
      case hexDataRecord:
         lowestAddress  = min(lowestAddress, addr + extendedAddress);
         highestAddress = max(lowestAddress, addr + extendedAddress + len - 1);
         bytesWritten += len;
      
         switch(action)
         {
            case checkFile:  // nothing much to do, we do the checks anyway
               break;
            
            case verifyFlash:
               verifyData(addr + extendedAddress, &hexBuffer[4], len);
               break;
            
            case writeToFlash:
               writeData(addr + extendedAddress, &hexBuffer[4], len);
               break;      
         } 
         break;
      
      // end of data
      case hexEndOfFile:
         gotEndOfFile = true;
         break;
      
      // we are setting the high-order byte of the address
      case hexExtendedSegmentAddressRecord: 
         extendedAddress = ((unsigned long) hexBuffer[4]) << 12;
         break;
      
      // ignore these, who cares?
      case hexStartSegmentAddressRecord:
      case hexExtendedLinearAddressRecord:
      case hexStartLinearAddressRecord:
         break;
      
      default:  
         ShowMessage(MSG_UNKNOWN_RECORD_TYPE);
         return true;  
   }  // end of switch on recType
   
   return false;
} // end of processLine


//------------------------------------------------------------------------------
// returns true if error, false if OK
bool readHexFile(const char * fName, const byte action)
{
   const int maxLine = 80;
   char buffer[maxLine];
   ifstream sdin(fName);
   int lineNumber = 0;
   gotEndOfFile = false;
   extendedAddress = 0;
   errors = 0;
   lowestAddress = 0xFFFFFFFF;
   highestAddress = 0;
   bytesWritten = 0;
   progressBarCount = 0;
   
   pagesize = currentSignature.pageSize;
   pagemask = ~(pagesize - 1);
   oldPage = NO_PAGE;
   
   // check for open error
   if(!sdin.is_open()) 
   {
      ShowMessage(MSG_CANNOT_OPEN_FILE);
      return true;
   }
   
   switch(action)
   {
      case checkFile:
         break;
      
      case verifyFlash:
         break;
      
      case writeToFlash:
         program(progamEnable, chipErase);   // erase it
         delay(20);  // for Atmega8
         pollUntilReady(); 
         break;      
   } 
   
   while(sdin.getline(buffer, maxLine))
   {
      lineNumber++;
      int count = sdin.gcount();
      if (sdin.fail()) 
      {
         ShowMessage(MSG_LINE_TOO_LONG);
         return true;
      } 
      
      // ignore empty lines
      if (count > 1)
      {
         if (processLine(buffer, action))
         {
            return true;  // error
         }
      }
   }    // end of while each line
   
   if(!gotEndOfFile)
   {
      ShowMessage(MSG_NO_END_OF_FILE_RECORD);
      return true;
   }
   
   switch(action)
   {
      case writeToFlash:
         // commit final page
         if(oldPage != NO_PAGE)
            commitPage(oldPage);
         break;
      
      case verifyFlash:
         if(errors > 0)
         {
            ShowMessage(MSG_VERIFICATION_ERROR);
            return true;
         }  
         break;
      
      case checkFile:
         break;
   }  // end of switch
   
   return false;
} 


// returns true if managed to enter programming mode
bool startProgramming()
{ 
   byte confirm;
   pinMode(RESET, OUTPUT);
   digitalWrite(MSPIM_SCK, LOW);
   pinMode(MSPIM_SCK, OUTPUT);
   pinMode(BB_MOSI, OUTPUT);
   unsigned int timeout = 0;
   
   // we are in sync if we get back programAcknowledge on the third byte
   do 
   {
      // regrouping pause
      delay(100);
      
      // ensure SCK low
      noInterrupts();
      digitalWrite(MSPIM_SCK, LOW);
      
      // then pulse reset, see page 309 of datasheet
      digitalWrite(RESET, HIGH);
      delayMicroseconds(10);  // pulse for at least 2 clock cycles
      digitalWrite(RESET, LOW);
      interrupts();
      
      delay(25);  // wait at least 20 mS
      noInterrupts();
      BB_SPITransfer(progamEnable);  
      BB_SPITransfer(programAcknowledge);  
      confirm = BB_SPITransfer(0);  
      BB_SPITransfer(0);  
      interrupts();
      
      if(confirm != programAcknowledge)
      {
         if(timeout++ >= ENTER_PROGRAMMING_ATTEMPTS)
            return false;
      }
   
   } while(confirm != programAcknowledge);

   // entered programming mode OK
   return true;  
}


void stopProgramming()
{
   // turn off pull-ups
   digitalWrite(RESET, LOW);  
   digitalWrite(MSPIM_SCK, LOW);
   digitalWrite(BB_MOSI, LOW);
   digitalWrite(BB_MISO, LOW);
   
   // set everything back to inputs
   pinMode(RESET, INPUT);
   pinMode(MSPIM_SCK, INPUT);
   pinMode(BB_MOSI, INPUT);
   pinMode(BB_MISO, INPUT);
}


void getSignature()
{
   foundSig = -1;
   lastAddressMSB = 0;
   
   byte sig[3];
   for(byte i = 0; i < 3; i++)
   {
      sig[i] = program(readSignatureByte, 0, i); 
   }  
      
   for(unsigned int j = 0; j < NUMITEMS(signatures); j++)
   {
      memcpy_P(&currentSignature, &signatures[j], sizeof currentSignature);
      
      if (memcmp(sig, currentSignature.sig, sizeof sig) == 0)
      {
         foundSig = j;
         
         // make sure extended address is zero to match lastAddressMSB variable
         program(loadExtendedAddressByte, 0, 0); 
         return;
      }  
   } 
   
   ShowMessage(MSG_UNRECOGNIZED_SIGNATURE);
} 


void getFuseBytes ()
{
   fuses[lowFuse]   = program(readLowFuseByte, readLowFuseByteArg2);
   fuses[highFuse]  = program(readHighFuseByte, readHighFuseByteArg2);
   fuses[extFuse]   = program(readExtendedFuseByte, readExtendedFuseByteArg2);
   fuses[lockByte]  = program(readLockByte, readLockByteArg2);
   fuses[calibrationByte]  = program(readCalibrationByte);
}


// write specified value to specified fuse/lock byte
void writeFuse(const byte newValue, const byte instruction)
{
   if(newValue == 0)
      return;  // ignore
   
   program(progamEnable, instruction, 0, newValue);
   pollUntilReady(); 
} 


// returns true if error, false if OK
bool updateFuses(const bool writeIt)
{
   unsigned long addr;
   unsigned int  len;
   
   byte fusenumber = currentSignature.fuseWithBootloaderSize;
   
   // if no fuse, can't change it
   if(fusenumber == NO_FUSE)
   {
      //    ShowMessage (MSG_NO_BOOTLOADER_FUSE);   // maybe this doesn't matter?
      return false;  // ok return
   }
   
   addr = currentSignature.flashSize;
   len = currentSignature.baseBootSize;
   
   if(lowestAddress == 0)
   {
      // don't use bootloader  
      fuses[fusenumber] |= 1;
   }
   else 
   {
      byte newval = 0xFF;
      
      if(lowestAddress == (addr - len))
         newval = 3;
      else if(lowestAddress == (addr - len * 2))
         newval = 2;
      else if(lowestAddress == (addr - len * 4))
         newval = 1;
      else if(lowestAddress == (addr - len * 8))
         newval = 0;
      else
      {
         ShowMessage(MSG_BAD_START_ADDRESS);
         return true;
      }
      
      if(newval != 0xFF)
      {
         newval <<= 1; 
         fuses[fusenumber] &= ~0x07;   // also program (clear) "boot into bootloader" bit 
         fuses[fusenumber] |= newval;  
      }
   }  
   
   if(writeIt)
   {
      writeFuse(fuses[fusenumber], fuseCommands[fusenumber]);      
   }
   
   return false;
}  


// returns true if OK, false on error
bool writeFlashContents()
{
   errors = 0;
      
   // ensure back in programming mode  
   if(!startProgramming())
      return false;
   
   // now commit to flash
   if(readHexFile(filename, writeToFlash))
      return false;
      
   // now fix up fuses so we can boot    
   if(errors == 0)
      updateFuses(true);
   
   return errors == 0;
} 



enum
{
   stateSleep,
   stateNoSDCard,
   stateMenu,
   stateAskEEPROMSave,
   stateAskEEPROMRestore,
   stateError,      
   stateUnknown,
} currStateType;
byte currState = stateSleep;
int stateStartTime = 0;


void setState(byte newState)
{
   currState = newState;   
   stateStartTime = millis();

   #ifdef DEBUG
      switch(newState)
      {
         case stateSleep:            DEBUGLN("stateSleep");            break;
         case stateNoSDCard:         DEBUGLN("stateNoSDCard");         break;
         case stateMenu:             DEBUGLN("stateMenu");             break;
         case stateAskEEPROMSave:    DEBUGLN("stateAskEEPROMSave");    break;
         case stateAskEEPROMRestore: DEBUGLN("stateAskEEPROMRestore"); break;
         case stateError:            DEBUGLN("stateError");            break;
         default:                    DEBUGLN("stateUnknown");          break;         
      }
   #endif
}

void releaseToArduboy()
{
   DEBUGLN("Releasing to arduboy");

   //make sure arduboy is in reset mode
   pinMode(RESET, OUTPUT);
   digitalWrite(RESET, LOW);
   
   //set all pins to input mode
   pinMode(BTN_UP,    INPUT);
   pinMode(BTN_DOWN,  INPUT);
   pinMode(BTN_LEFT,  INPUT);
   pinMode(BTN_RIGHT, INPUT);
   pinMode(BTN_IN,    INPUT);
   pinMode(BTN_OUT,   INPUT);
   pinMode(BTN_MENU,  INPUT);

   pinMode(PIN_SD_CD, INPUT);

   pinMode(MSPIM_SCK, INPUT);
   pinMode(MSPIM_SS,  INPUT);
   pinMode(BB_MISO,   INPUT);
   pinMode(BB_MOSI,   INPUT);

   pinMode(PIN_OLED_RES, INPUT);
   pinMode(PIN_OLED_CS,  INPUT);
   pinMode(PIN_OLED_DC,  INPUT);
   pinMode(13,           INPUT);  //SCK
   pinMode(11,           INPUT); //MOSI

   //arduboy is multiplex channel A
   // pin high = B, pin low = A
   pinMode(PIN_OLED_PLEX, OUTPUT);
   digitalWrite(PIN_OLED_PLEX, LOW);

   //but we need to listen to menu button
   // update: we moved it to the ADC-only pins, which are always input
   //pinMode(BTN_MENU, INPUT_PULLUP);
   //digitalWrite(BTN_MENU, HIGH);

   //let everything settle
   delay(100);

   //un-reset arduboy
   digitalWrite(RESET, HIGH);

   setState(stateSleep);
}


void wakeUp()
{
   DEBUGLN("Waking up!");
         
   //set target reset pin to output
   pinMode(RESET, OUTPUT);
   
   //hold arduboy in reset state
   digitalWrite(RESET, LOW);

   //switch display to icsp core
   digitalWrite(PIN_OLED_PLEX, HIGH);

   //wait a sec to be sure that arduboy is in reset state
   delay(100);
     
   //init button pins to input_pullup
   // pressing the buttons connects them to GND
   pinMode(BTN_UP,    INPUT_PULLUP); 
   pinMode(BTN_DOWN,  INPUT_PULLUP); 
   pinMode(BTN_LEFT,  INPUT_PULLUP); 
   pinMode(BTN_RIGHT, INPUT_PULLUP); 
   pinMode(BTN_IN,    INPUT_PULLUP); 
   pinMode(BTN_OUT,   INPUT_PULLUP); 
   digitalWrite(BTN_UP,    HIGH);
   digitalWrite(BTN_DOWN,  HIGH);
   digitalWrite(BTN_LEFT,  HIGH);
   digitalWrite(BTN_RIGHT, HIGH);
   digitalWrite(BTN_IN,    HIGH);
   digitalWrite(BTN_OUT,   HIGH);

   //clear out any floating pin values
   updateButtonState();
   updateButtonState();
      
   initSDCard();   

   //reset/initialize display
   oled.init();
   //oled.begin(&Adafruit128x64, PIN_OLED_CS, PIN_OLED_DC);
   //oled.setFont(Wendy3x5);
   oled.clear();                 // clear screen
   setState(stateMenu);
   readListLine();
}


void initSDCard()
{
   //init card detect pin
   pinMode(PIN_SD_CD, INPUT);

   //check SD card detect pin
   // it goes to 3.3v when card is inserted
   // might want to use an analog pin for this?
   if(analogRead(PIN_SD_CD) > ANALOG_THRESHOLD) //digitalRead(PIN_SD_CD) == HIGH)
   {
      //card inserted, attempt to intialize card
      if(sd.begin(chipSelect, SPI_QUARTER_SPEED)) 
      {
         if(listFileStream.is_open())
            listFileStream.close();
         listFileStream.open("list.txt");

         if(listLineCount <= 0)
            countListLines();
         
         //readListLine();
         
         //sd.ls(LS_R);

         //Serial.println(listLineBuff);   
         
         setState(stateMenu);   
         
      }
      else
      {
         ShowMessage(MSG_SD_CARD_INIT_FAILED); //since we have the card detect, this should be "failed ot initialize", not "no sd card"
         setState(stateError);
      }      
   }
   else
   {
      //card not inserted
      setState(stateNoSDCard);
   }   
}


void countListLines()
{
   listFileStream.seekg(0);

   while(!listFileStream.eof())
   {
      listFileStream.getline(listLineBuff, listLineSize);
      listLineCount++;
   }
}

void readListLine()
{
   //update list selection
   // and display menu

   //screen can fit 8 lines of text
   // set an upper and lower range for us to print
   int menuStart = listLineIdx - 3;
   if(menuStart < 0) 
      menuStart = 0;
   int menuEnd = menuStart + 7;
   if(menuEnd >= listLineCount)
   {
      menuEnd = listLineCount - 1;
      menuStart = menuEnd - 7;
      if(menuStart < 0)
         menuStart = 0;
   }

   //clear display
   DEBUGLN("---------------");
   #ifdef DEBUG
      Serial.print(menuStart);
      Serial.print(" to ");
      Serial.println(menuEnd);
   #endif
   //oled.clear();                 // clear screen
   oled.setCursor(0,0);
   
   for(int i = menuStart; i <= menuEnd; i++)
   {
      listFileStream.seekg(i * listLineSize);
      if(listFileStream.fail())
      {
         listLineIdx = 0;
         listFileStream.seekg(0);
         if(listFileStream.fail())
         {
            DEBUGLN("Error: can't seek list.txt");
            return;
         }
         DEBUGLN("listFileStream seek error");
         setState(stateError);
         return;         
      }

      listFileStream.getline(listLineBuff, listLineSize);   

      //remove trailing spaces
      for(int i = listLineSize; i > 23 ; i--)
      {
         //if(listLineBuff[i] != 0 && listLineBuff[i] != '_')
         //   break;

         //if(listLineBuff[i] == '_')
         //   listLineBuff[i] = ' ';       

         if(listLineBuff[i] == ' ' || listLineBuff[i] == 0)
            listLineBuff[i] = 0;      
         else
            break;                         
      }

      //if this is the selected file, extract filename
      if(i == listLineIdx)
      {
         //extract filename
         strncpy(filename, listLineBuff, 9);
         //append .hex extension
         for(int i = 0; i <= 9; i++)
         {
            if(filename[i] != ' ') //8.3 filenames can't have spaces, so this is used as delimiter
               continue;
            filename[i+0] = '.';               
            filename[i+1] = 's';
            filename[i+2] = 'h';
            filename[i+3] = 't';
            filename[i+4] = 0;
         }

         //also add cursor to display name
         #ifdef DEBUG
            listLineBuff[8] = '>';
         #endif
         oled.inverted = 1;
         oled.sendData(0xFF);
      }
      
      //print the menu entry
      #ifdef DEBUG
         Serial.println(&listLineBuff[8]);         
      #endif

      oled.writeWidth(&listLineBuff[9], 64);
      oled.writeWidth("\n", 0);
      oled.inverted = false;
   }

   //display a screenshot
   ifstream shot(filename);
   byte b = 0;
   for(int i = 0; i < 8; i++)
   {
      oled.setCursor(i, 64);
      for(int j = 0; j < 64; j++)
      {  
         b = shot.get();
         oled.sendData(b);         
      }
   }

   //change extension back to .hex for loading
   for(int i = 0; i <= 9; i++)
   {
      if(filename[i] != '.')
         continue;
      filename[i+1] = 'h';
      filename[i+2] = 'e';
      filename[i+3] = 'x';
      filename[i+4] = 0;
   }
}


void updateButtonState()
{
   prevButtonState = currButtonState;
   currButtonState = 0;

   if(digitalRead(BTN_UP)    == LOW) currButtonState |= bit(BTN_UP);
   if(digitalRead(BTN_DOWN)  == LOW) currButtonState |= bit(BTN_DOWN);
   if(digitalRead(BTN_LEFT)  == LOW) currButtonState |= bit(BTN_LEFT);
   if(digitalRead(BTN_RIGHT) == LOW) currButtonState |= bit(BTN_RIGHT);
   if(digitalRead(BTN_IN)    == LOW) currButtonState |= bit(BTN_IN);
   if(digitalRead(BTN_OUT)   == LOW) currButtonState |= bit(BTN_OUT);
   //if(digitalRead(BTN_MENU)  == LOW) currButtonState |= bit(BTN_MENU);
   if(analogRead(BTN_MENU) < ANALOG_THRESHOLD) currButtonState |= bit(BTN_MENU);

   //Serial.println(bit(31) | currButtonState, BIN);
   //delay(200);
}


bool justPressed(int pin)
{
   /*
   Serial.println("justPressed()");
   Serial.println(prevButtonState, BIN);
   Serial.println(currButtonState, BIN);
   Serial.println("");
   */
   
   if((currButtonState & bit(pin)) == 0)   
      return false;
   if((prevButtonState & bit(pin)) > 0)
      return false;

   return true;     
}


bool justReleased(int pin)
{
   if((currButtonState & bit(pin)) > 0)   
      return false;
   if((prevButtonState & bit(pin)) == 0)
      return false;

   return true;     
}


void loadSelectedGame()
{
   //flashes whatever hex file we currently have selected to the arduboy cor

   oled.clear();
   oled.setCursor(0,0);
   oled.writeWidth("Loading");
   delay(100);

   if(!startProgramming())
   {
      ShowMessage(MSG_CANNOT_ENTER_PROGRAMMING_MODE);

      //something went wrong, keep the arduboy reset
      digitalWrite(RESET, LOW);
      setState(stateError);
            
      return;
   }  
      
   getSignature();
   getFuseBytes();
   
   // don't have signature? don't proceed
   if (foundSig == -1)
   {
      ShowMessage(MSG_CANNOT_FIND_SIGNATURE);

      //something went wrong, keep the arduboy reset
      digitalWrite(RESET, LOW);
      setState(stateError);
      
      return;
   }
   
   bool ok = writeFlashContents();
   stopProgramming();
   delay(500);
   
   if(ok)
   {
      DEBUGLN("Flashed OK!");
      releaseToArduboy();
   }
   else
   {
      //something went wrong, keep the arduboy reset
      digitalWrite(RESET, LOW);
      setState(stateError);
   }
}

void dumpEEPROM()
{
   //read all the bytes from the arduboy's eeprom and print them to serial
   if(!startProgramming())
   {
      ShowMessage(MSG_CANNOT_ENTER_PROGRAMMING_MODE);

      //something went wrong, keep the arduboy reset
      digitalWrite(RESET, LOW);
      setState(stateError);
            
      return;
   }  

   for(int i = 0; i < 512; i++)
   {
      byte highByte = (i >> 8) & 0x01;
      byte lowByte = i & 0xFF;
      byte value = program(readEEPROMMemory, highByte, lowByte, 0x00);
      Serial.print(value, HEX);
      if(i % 64 == 0)
         Serial.println();      
   }

   stopProgramming();
   delay(500);      
}

void setup() 
{
   releaseToArduboy();

   //reset arduboy
   pinMode(RESET, OUTPUT);
   digitalWrite(RESET, LOW);

   //display multiplexer
   pinMode(PIN_OLED_PLEX, OUTPUT);
   digitalWrite(PIN_OLED_PLEX, LOW);

   #ifdef DEBUG
   {
      Serial.begin(57600);
      delay(100);
      DEBUGLN("Serial Online");
   }
   #else
      //wait a sec to be sure that arduboy is in reset state
      delay(10);
   #endif

   digitalWrite(RESET, HIGH);

   //delay(100);
   //wakeUp();
   
}


void loop() 
{
   int elapsedTime = millis() - stateStartTime;
   
   switch(currState)
   {
      case stateSleep:
         //if we press the menu button, wake up
         #ifdef DEBUG
            Serial.print(analogRead(BTN_MENU));
            Serial.print(" ");
            Serial.print(analogRead(PIN_SD_CD));            
            Serial.println(" sleep");
            delay(50);
         #endif
         
         updateButtonState();
         if(justPressed(BTN_MENU))
         {
            wakeUp();
         }

         //otherwise conserve power?
         break;

      case stateNoSDCard:
         //try to re-init when we insert a card
         if(analogRead(PIN_SD_CD) > ANALOG_THRESHOLD) //digitalRead(PIN_SD_CD) == HIGH)
            initSDCard();
         
         //blink message on screen
         if(elapsedTime > 500)
         {
            DEBUGLN("Insert SD Card");
            setState(stateNoSDCard);
         }

         //A = return to arduboy
         updateButtonState();
         if(justPressed(BTN_OUT))
         {
            releaseToArduboy();
         }
         break;

      case stateMenu:
         //up/down = prev/next game
         updateButtonState();
         
         if(justPressed(BTN_MENU))
         {
            wakeUp();
         }
         
         if(justPressed(BTN_UP))
         {
            listLineIdx--;
            if(listLineIdx < 0)
               listLineIdx = listLineCount - 1;
               
            readListLine();            
         }
         else if(justPressed(BTN_DOWN))
         {
            listLineIdx++;
            if(listLineIdx >= listLineCount)
               listLineIdx = 0;
               
            readListLine();
         }

         //A = load game
         if(justPressed(BTN_OUT))
         {
            Serial.print(F("Load Game "));
            Serial.println(filename);

            loadSelectedGame();
         }

         //if sd card removed, go to no sd card state
         if(analogRead(PIN_SD_CD) < ANALOG_THRESHOLD) //digitalRead(PIN_SD_CD) == LOW)
         {
            setState(stateNoSDCard);
         }
         
         break;

      case stateAskEEPROMSave:
         //do you want to save current eeprom? yes/no
         break;

      case stateAskEEPROMRestore:
         //do you want to restore saved eeprom? yes/no
         break;

      case stateError:
         //error message is displayed on screen
         
         //if we pull the sd card out, go back to no-sd card state
         // sometimes error can be triggered because the card is disconnected but card detect hasnt updated yet
         if(analogRead(PIN_SD_CD) < ANALOG_THRESHOLD) //digitalRead(PIN_SD_CD) == LOW)
         {
            setState(stateNoSDCard);
         }

         updateButtonState();
            
         //B => go back to menu
         if(justPressed(BTN_IN))
         {
            setState(stateNoSDCard);
         }
         
         if(justPressed(BTN_MENU))
         {
            wakeUp();
         }
         
         //A => return to last arduboy game
         if(justPressed(BTN_OUT))
         {
            releaseToArduboy();
         }
         break;

      default:
         DEBUGLN("Error: Unknown state");         
         break;
   }
}
